using System;
using System.Threading;

// Eksempel med brug af guarded suspension

namespace ObserverPattern_Parkeringshus
{
    public class P_Hus_Monitor
    {
        private int antalBiler; // biler i huset
        private int pladser;	// antal parkeringspladser
        private object phus_ls = new object();     // privat lseobjekt

        public delegate void EventHandler (string besked, P_Hus_Monitor p_hus);
        public event EventHandler P_Hus_EventAlle;
        
        public event EventHandler P_Hus_EventInd;
        public event EventHandler P_Hus_EventIndStart;
        public event EventHandler P_Hus_EventIndSlut;
        public event EventHandler P_Hus_EventIndVenter;
        
        public event EventHandler P_Hus_EventUd;
        public event EventHandler P_Hus_EventUdStart;
        public event EventHandler P_Hus_EventUdSlut;
        public event EventHandler P_Hus_EventUdVenter;

        public event EventHandler P_Hus_EventFyldt;
        public event EventHandler P_Hus_EventTmt;
        public event EventHandler P_Hus_EventIkkeFyldt;
        public event EventHandler P_Hus_EventIkkeTomt;


        public P_Hus_Monitor (int pladser)
        {
            this.pladser = pladser;
            this.antalBiler = 0;

            this.P_Hus_EventInd += this.aktivet_P_Hus_EventAlle;        // aktivitet_P_Hus_EventAlle er observer p EventInd
            this.P_Hus_EventUd += this.aktivet_P_Hus_EventAlle;         // aktivitet_P_Hus_EventAlle er observer p EventInd

            this.P_Hus_EventIndStart += this.aktivet_P_Hus_EventInd;    // aktivet_P_Hus_EventInd er observer p EventIndStart
            this.P_Hus_EventIndSlut += this.aktivet_P_Hus_EventInd;     // aktivet_P_Hus_EventInd er observer p EventIndSlut
            this.P_Hus_EventIndVenter += this.aktivet_P_Hus_EventInd;   // aktivet_P_Hus_EventInd er observer p EventIndSlut
            this.P_Hus_EventFyldt += this.aktivet_P_Hus_EventInd;       // aktivet_P_Hus_EventInd er observer p EventIndSlut
            this.P_Hus_EventIkkeTomt += this.aktivet_P_Hus_EventInd;    // aktivet_P_Hus_EventInd er observer p EventIndSlut

            this.P_Hus_EventUdStart += aktivet_P_Hus_EventUd;          // aktivet_P_Hus_EventUd er observer p EventIkkeFyldt
            this.P_Hus_EventUdSlut += aktivet_P_Hus_EventUd;           // aktivet_P_Hus_EventUd er observer p EventIkkeFyldt
            this.P_Hus_EventUdVenter += aktivet_P_Hus_EventUd;         // aktivet_P_Hus_EventUd er observer p EventIkkeFyldt
            this.P_Hus_EventTmt += aktivet_P_Hus_EventUd;             // aktivet_P_Hus_EventUd er observer p EventIkkeFyldt
            this.P_Hus_EventIkkeFyldt += aktivet_P_Hus_EventUd;        // aktivet_P_Hus_EventUd er observer p EventIkkeFyldt
        }

        private void aktivet_P_Hus_EventAlle(string besked, P_Hus_Monitor p_hus)
        {
            if (P_Hus_EventAlle != null)
                P_Hus_EventAlle(besked, p_hus);      // "vidre-send" til observere p P_Hus_EventAlle
        }
        private void aktivet_P_Hus_EventInd(string besked, P_Hus_Monitor p_hus)
        {
            if (P_Hus_EventInd != null)
                P_Hus_EventInd(besked, p_hus);      // "vidre-send" til observere p P_Hus_EventInd
        }
        private void aktivet_P_Hus_EventUd(string besked, P_Hus_Monitor p_hus)
        {
            if (P_Hus_EventUd != null)
                P_Hus_EventUd(besked, p_hus);      // "vidre-send" til observere p P_Hus_EventUd
        }

        public void bilKrerInd(int nr, int count)
        {

            P_Hus_EventIndStart("Start p bilKrerInd :" + +nr + ":" + count, this);
            lock (phus_ls)
            {
                P_Hus_EventInd("Start p bilKrerInd - Enter kritisk omrde:" + +nr + ":" + count, this);
                while (antalBiler == pladser)
                {
                    P_Hus_EventIndVenter("Parkeringshus fuldt - ingen indkrsler nu:" + +nr + ":" + count, this);
                    Monitor.Wait(phus_ls);
                }
                if (antalBiler == 0)
                {
                    P_Hus_EventIkkeTomt("ben for udkrsler igen - 1. bil lige krt ind:" + +nr + ":" + count, this);
                    Monitor.PulseAll(phus_ls);
                }
                antalBiler++;
                P_Hus_EventIndSlut("Bil krt ind - der er nu ialt:" + antalBiler + " biler:" + +nr + ":" + count, this);
                if (antalBiler == pladser)
                    P_Hus_EventFyldt("P-hus er fyldt:" + +nr + ":" + count, this);
            }
        }

        public void bilKrerUd(int nr, int count)
        {
            P_Hus_EventUdStart("Start p bilKrerUd:" + +nr + ":" + count, this);
            lock (phus_ls)
            {
                P_Hus_EventUd("Start p bilKrerUd - Enter kritisk omrde:" + +nr + ":" + count, this);
                while (antalBiler == 0)
                {
                    P_Hus_EventUdVenter("Parkeringshus tomt - ingen udkrsler nu:" + +nr + ":" + count, this);
                    Monitor.Wait(phus_ls);
                }
                if (antalBiler == pladser)
                {
                    P_Hus_EventIkkeFyldt("ben for indkrsler igen - sidste plads bliver ledig:" + +nr + ":" + count, this);
                    Monitor.PulseAll(phus_ls);
                }
                antalBiler--;
                P_Hus_EventUdSlut("Bil krer ud - der er nu ialt:" + antalBiler + " biler:" + +nr + ":" + count, this);
                if (antalBiler == 0)
                    P_Hus_EventTmt("P-hus er tmt", this);
            }
        }
        public bool Ls()
        {
            return Monitor.TryEnter(phus_ls);      // true hvis ls opnet
        }
        public void LsOp()
        {
            Monitor.Exit(phus_ls);
        }
    }
}
